/*
 *	Copyright 2002, Lucent Technologies.
 *      Author:  Bill Tanenbaum		      
 *
 *	libvp.c:
 *
 *	cgrep version of the viewpath library:
 *
 *	Uses NVPATH, which it reads from the environment (getenv()).
 *
 *	Supports viewpathed open() and stat() system calls only.
 *
 *	NOTE: The viewpathed calls are vpopen() and vpstat(),
 *	not open() and stat(), so this version is not fully
 *	transparent to the caller, unlike the separate archive
 *	viewpath libraries.
 *
 */

#define VPATH	"NVPATH"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <dirent.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>

#define function		/* fake keyword to make it easy to index */
				/* thru functions while in the editor.	 */
#define READ 0
#define BLOCK 1024
#define DIRLEN 255
#define MAXDIR 32		/* maximum number of nodes in a viewpath */
#define MAXDIRS 32		/* maximum number of directories in a path */
#define FAILCODE	-1	/* stat fail indicator for non-file */

extern int errno;		/* location of error number from system call */

extern int vpopen();
extern int vpstat();

extern int read();
extern int write();

/*
 *	Data for the viewpath library:
 */

static char V_fname[256];

static int V_flag = 1;		/* flag to turn viewpathing on and off	*/
				/* viewpathing starts out on by default */

static int V_wrstat = 0;	/* by default, return the real status	*/
				/* of the file as found down the        */
				/* viewpath				*/

static int V_changed = 1;	/* flag which may be used to tell that */
				/* the viewpath has changed.  This will */
				/* someday be used so the shell can	*/
				/* tell the viewpath library that the 	*/
				/* viewpath was changed.  Currently	*/
				/* we assume it may have for the shell 	*/
				/* and ksh versions of the viewpath	*/
				/* library.  It may be used effectively	*/
				/* for other programs which wish to 	*/
				/* alter the value of VPATH in their 	*/
				/* environment.				*/

static char V_pathname[MAXPATHLEN];
				/* common static pathname used for 	*/
				/* searching down the viewpath in open,	*/
				/* stat, access, and all the other 	*/
				/* viewpathed system calls.		*/

static char V_cwd[MAXPATHLEN];	/* full pathname of the current working	*/
				/* directory.  Set by V_init().		*/

static int V_did_chdir = 1;	/* flag set by chdir() to tell V_init()	*/
				/* that it must call getcwd() to find	*/
				/* the new working directory.		*/

static char *V_nodes[MAXDIR];	/* The nodes in the viewpath.		*/

static char Viewpath[BLOCK] = "";
				/* space to copy and split the viewpath	*/
				/* so that its pieces may be stored in	*/
				/* V_nodes (above).			*/

static char V_relpath[MAXPATHLEN];
				/* used to store the filename relative	*/
				/* to the current node while searching	*/
				/* for a file down the viewpath.	*/

static int V_ndir = 0;		/* the number of nodes in the viewpath.	*/

static int V_node = 0;		/* index of node file was in (for stat) */


extern char *getcwd();
/*--------------------------------------------------------------------------
 *
 *	Viewpath library utility Functions.
 *	Note that since the viewpath library must be as self-contained
 *	as possible, some unix utility Functions are virtually
 *	here by different names.  This is to avoid clashes with
 *	library Functions redefined by programs linked with a viewpath
 *	library.
 *
 *--------------------------------------------------------------------------
 *
 *	void V_cat(dest, source1, ..., 0)
 *	char *dest, *source1, ...;
 *
 *	Description:
 *		Concatenate an arbitrary number of strings, into a
 *		pre-allocated destination string.  It is more general
 *		than strcat(), and avoids the hazards of bringing in
 *		C library functions which might clash with functions
 *		in programs linked with a viewpath library (i.e., we
 *		might end up calling somebody elses strcat() function,
 *		which takes different parameters than the standard unix
 *		version).
 *
 *--------------------------------------------------------------------------
 *
 *	int V_init()
 *
 *	Description:
 *		Function to make sure the viewpath, and current working
 *		directory are up to date.  It initializes V_ndir (the number
 *		nodes in the viewpath), V_nodes (the nodes in the viewpath),
 *		and V_cwd.  It should be called by each
 *		viewpathed system call prior to looking for a file down
 *		the viewpath.
 *
 *--------------------------------------------------------------------------
 *
 *	void V_compath(path)
 *	char *path;
 *
 *	Description:
 *		Compress a pathname, using the following rules:
 *
 *		+ Replace '//' with '/'.
 *		+ Remove directories named '.' from pathname.
 *		+ Remove '<x>/..' from pathname.
 *		+ Remove trailing '/'.
 *		+ We don't bother removing leading '/..'s.
 *		+ If we eliminate everything, put back one '.'.
 *
 *		The algorithm used involves compressing the pathname into
 *		itself, with a 'read' and 'write' pointer (note that the
 *		write pointer can't pass the read pointer, since we never
 *		make the path longer).
 *
 *--------------------------------------------------------------------------
 *
 *	char *V_getrel(filename)
 *	char *filename;
 *
 *	Description:
 *		Put the path to filename relative to the top of the current
 *		node in V_relpath (and return it), if filename is in the
 *		current node. Filename may be relative or absolute, and we
 *		may be outside the node, giving a path into the node.
 *
 *--------------------------------------------------------------------------
 */

static void V_get_vpath();

static
void
V_cat(char *dest, ...)
{
	va_list sp;
	register char *d = dest;
	register char *s;

	va_start(sp, dest);

	d = dest;
	while (NULL != (s = va_arg(sp, char *))) {
		while (*d++ = *s++) ;
		d--;
	}
	
	va_end(sp);
} /* end of V_cat() */

static
void
V_get_vpath()
{
	register char *vptr;
#ifdef uts
	char *getenv();
#endif
	if (!V_flag) {
		return;
	}
	
	if (V_changed) {
		V_changed = 0;
		vptr = getenv(VPATH);
		
		if ((vptr == NULL) || (*vptr == '\0')) {
			V_ndir = 0;
			V_cat(Viewpath, "", 0);
			V_cat(V_relpath, "", 0);
			return;
		}
		
		V_cat(Viewpath, vptr, 0);
		vptr = Viewpath;
		V_ndir = 0;
	
		while (*vptr){
#ifndef uts
			char resolv_path[BLOCK];
			char *tpath;
#ifdef sparc
			char *realpath();
#endif

			V_nodes[V_ndir] = vptr;
#else
			V_nodes[V_ndir++] = vptr;
#endif
			while (*vptr && (*vptr != ':')){
				vptr++;
			}
			if (*vptr){
				*vptr = '\0';
				vptr++;
			}
#ifndef uts

			/* This code corrects the problem of comparing the
			   realpath returned by getcwd with any symbolically
			   linked path in the VPATH. The problem could
			   potentially arise on the MAXIs but has not been
			   seen so far so will not be addressed at this time.
			*/
			if (realpath(V_nodes[V_ndir],resolv_path) != (char *)NULL)
			{
				if ((tpath = (char *)malloc((unsigned)strlen(resolv_path) + 1)) != (char *)NULL)
				{
					(void)strcpy(tpath,resolv_path);
					V_nodes[V_ndir++] = tpath;
				}
				else
					perror("ERROR: libvp: V_get_vpath:  cannot malloc");
			}
#endif
		}
	}
} /* end of V_get_vpath() */

static
int
V_init()
{
	char *l_cwd;
	int   save_errno;
	save_errno = errno;
	V_get_vpath();
	
	if (V_did_chdir && V_ndir > 0) {
		l_cwd = getcwd((char *)NULL,256);
		V_cat(V_cwd,l_cwd,0);

		V_did_chdir = 0;
	}
	errno = save_errno;

	return(V_ndir);
} /* end of V_init() */

static
void
V_compath(pathname)
char *pathname;
{
	static char *components[MAXDIRS];   /* pointers to component dirs */
	static int isdotdot[MAXDIRS];	    /* boolean array for '..'s */

	register char *rptr = pathname;		/* read pointer */
	register char *wptr = pathname;		/* write pointer */
	int dirptr = 0;			/* index into components */

	/*
	 *	A null pathname is assumed to mean the current directory.
	 *	We insert "." into pathname, and return.
	 */
	
	if (*pathname == '\0') {
		pathname[0] = '.';
		pathname[1] = '\0';
		return;
	}

	/*
	 *	If the pathname begins with a '/', save it, and increment
	 *	the pointer to the first component directory name.
	 *	In any case, increment the read and write pointers, since
	 *	we will always want to copy the first character.
	 *	Note that this makes the loop condition below simpler,
	 *	and handles the special case of a leading '/'.
	 */
	
	*components = pathname;
	if (*pathname == '/') {
		++*components;
	}
	++rptr;
	++wptr;

	/*
	 *	Loop through the pathname (with rptr), copying information
	 *	to wptr, backing up or not incrementing wptr where
	 *	appropriate to compress the path.
	 *	We test to see if the next to last character is zero, so 
	 *	we can treat the end of pathname ('\0') just like a '/'.
	 */
	
	while (rptr[-1]) {

		if ((*rptr == '/') || (*rptr == '\0')) {
			/*
			 *	We have just found the end of one component
			 *	of the pathname.  Check to see if it is an 
			 *	extra '/', a '.', or a '..'.
			 *	In any case, increment rptr, and continue.
			 */

			if (wptr[-1] == '/') {	/* eliminate extra '/' */
				++rptr;		/* Note that this works ok */
				continue;	/* for trailing '/' case. */
			}

			if ((wptr[-1] == '.')	/* was it '..'? */
				&& ((wptr-2 == pathname) || (wptr[-3] == '/'))
				&& (wptr[-2] == '.')) {

				isdotdot[dirptr] = 1;	/* yes, it was */
			} else {
				isdotdot[dirptr] = 0;	/* no, it wasn't */
			}
			
			if ((dirptr >= 1)	/* was it '..' and is it   */
				&& isdotdot[dirptr]	/* after something */
				&& ! isdotdot[dirptr-1]) {	/* else ? */
					/*
					 *	If so, eliminate <dir>/..
					 */
					
				wptr = components[--dirptr];
				++rptr;
				continue;
			}
			
			if (((wptr-1 == pathname) || (wptr[-2] == '/'))
				&& (wptr[-1] == '.')) {
					/*
					 *	It was a '.'! get rid of it.
					 */
					
				++rptr; --wptr;
				continue;
			}

			/*
			 *	Mark start of next component
			 */
			
			*wptr++ = *rptr++;
			components[++dirptr] = wptr;
			continue;
		}

		/*
		 *	Otherwise, if it wasn't the end of a path element,
		 *	just copy it from rptr to wptr.
		 */
		
		*wptr++ = *rptr++;
	} /* end while loop to compress pathname */

	/*
	 *	Make sure we terminated pathname with a null.  We might have
	 *	already since we looped until one after the null,
	 *	but if the last component was '..', we didn't.
	 */
	
	*wptr = '\0';
	
	/*
	 *	If we eliminated everything, as we would for 'x/y/../..',
	 *	Put back one '.':
	 */
	
	if (wptr == pathname) {
		pathname[0] = '.';
		pathname[1] = '\0';
		return;
	}
	
	/*
	 *	Remove trailing '/', if any.  Be careful not to remove 
	 *	'/' if it is the only thing left.
	 */
	
	--wptr;
	if ((wptr > pathname) && (*wptr == '/')) {
		*wptr = '\0';
	}
} /* end of V_compath() */

static
char *
V_getrel(filename)
char *filename;
{
	register char *node;
	register char *path;
	
	node = V_nodes[0];
	
	if (*filename == '/') {
		path = filename;
	} else {
		V_cat(V_pathname, V_cwd, "/", filename, 0);
		V_compath(V_pathname);
		path = V_pathname;
	}

	while ((*node != '\0') && (*path == *node)) {
		++path;
		++node;
	}
	
	if (*node != '\0') {
		V_cat(V_relpath, "", 0);
		return(V_relpath);
	}
	
	if (*path == '\0') {
		V_cat(V_relpath, ".", 0);
		return(V_relpath);
	}
	
	if (*path != '/') {
		V_cat(V_relpath, "", 0);
		return(V_relpath);
	}
	
	while (*path == '/') {
		++path;
	}
	
	if (*path == '\0') {
		V_cat(V_relpath, ".", 0);
		return(V_relpath);
	}
	
	V_cat(V_relpath, path, 0);
	return(V_relpath);
	
} /* end of V_getrel() */

/************************************************************************
 *									*
 *	Fully viewpathed open() system call.				*
 *									*
 *	In the table below, the action associated with the first 	*
 *	condition which matches is taken.				*
 *									*
 *	CONDITION		ACTION					*
 *									*
 *	path begins	substitute root node for ^			*
 *	with a ^							*
 *									*
 *	path points	Pass on to standard open().			*
 *	out of ROOT							*
 *	node or !V_flag							*
 *									*
 ************************************************************************/

function int vpopen(filename, type)
char *filename;		/* name of file for open */
int type;		/* access type for file to be opened */
{
	register int  opencode;		/* result of attempted open */
	int pathnumber;		/* index into viewpath array */

	if ((filename != (char *)NULL) &&
            (filename[0] == '^') &&
            (filename[1] == '/' || filename[1] == '\0')) {
		if (0 != V_init()) {
			V_cat(V_fname, V_nodes[0], filename + 1, 0);
			filename = V_fname;
		}
	}
	
	opencode = open(filename, type, 0);

	if(!V_flag
	    || (opencode != FAILCODE)
	    || (type != READ)
	    || (0 == V_init())
	    || (*V_getrel(filename) == '\0')) {
		goto open_done;
	}

	/* open file using successive paths in VPATH */

	for (pathnumber = 1; pathnumber < V_ndir; pathnumber++) {
		V_cat(V_pathname, V_nodes[pathnumber], "/", V_relpath, 0);
		opencode = open(V_pathname, type, 0);

		if (opencode != FAILCODE) {
			goto open_done;
		}
	}
open_done:
	return(opencode);
} /* end of open() */

/************************************************************************
 *									*
 *	Viewpathed stat system call.					*
 *									*
 *	If the file is not in the current node, viewpathing is on, 	*
 *	VPATH is set, look for it down the viewpath.			*
 *									*
 ************************************************************************/

function vpstat(filename, buffer)
char *filename;			/* name of file for stat */
struct stat *buffer;		/* information buffer for stat */
{
	int statcode;		/* result of attempted stat */

	if ((filename != (char *)NULL) &&
            (filename[0] == '^') &&
            (filename[1] == '/' || filename[1] == '\0')) {
		if (0 != V_init()) {
			V_cat(V_fname, V_nodes[0], filename + 1, 0);
			filename = V_fname;
		}
	}
	
	V_node = 0;
	
	statcode = stat(filename, buffer);
	if ((!V_flag)
	  || (statcode != FAILCODE)
	  || (0 == V_init())
	  || (*V_getrel(filename) == '\0')) {
		return(statcode);
	}
	
	/*
	 *	V_relpath now contains the path to the file relative to the 
	 *	top of the node.  Search for the file in this position
	 *	in the other nodes. If the file is found and the V_wrstat
	 *      flag is set, return a permission that indicates that the
	 * 	file is writable. If V_wrstat is defaulted to 0, the actual
	 *	permission of the file is returned.
	 */
	
	for (V_node = 1; V_node < V_ndir; V_node++) {
		V_cat(V_pathname, V_nodes[V_node], "/", V_relpath, 0);
		statcode = stat(V_pathname, buffer);
		if (statcode >= 0) {
			if (V_wrstat)
				buffer->st_mode |= 0222;
			break;
		}
	}

	return(statcode);
} /* end of stat() */
