	static char sccsid[]="%Z%	match.c	%E% %U%";
/******************************************************************************

                ###############################################
                # Copyright 1987-2000 Lucent Technologies Inc #
		# Author:  Bill Tanenbaum		      #
                ###############################################

Module:	match.c [%P%]

Description:

	General purpose pattern matching routines.

Functions:
	pattern_match()
	vfast_reverse()
	store_it()
	re_lnexec()
	re_lbexec()
	re_nullexec()
	fixloc()
	grepmatch()
	check_count()
	match_alloc()

*******************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include	"re.h"

static CHAR *S1, *F1;

static BOOL fixloc(), check_count();
static int vfast_reverse(), store_it();

/******************************************************************************

Routine:
	pattern_match()

Syntax:

	int pattern_match(pat, begin, end, nsaved)

	PATTERN *pat;	// Pointer to pattern to match
	CHAR	*begin;	// Pointer to first character of text to be searched
	CHAR	*end;	// Pointer to last character of text to be searched
	int	nsaved;	// Number of stored matches left over from the previous
			//	text buffer that must be saved.  Note that these
			//	matches must all be BEFORE "begin".  None can be
			//	within the "begin" to "end" range.
Description:

	Searches a specified range of text for matches to a specific pattern,
	and stores the matches.

	pat->N_start is the number of matches to skip over without storing.
	pat->N_start is decremented as each match is skipped.

	pat->N_stop, if non-zero, is the maximum number of matches to find
	(after skipping over pat->N_start matches).  If zero, there is no upper
	limit on the number of matches found.  If non-zero, pat->N_stop is
	decremented as each match is found.  If pat->N_stop is decremented to zero,
	pat->M_done is set, indicating no more matches.

	pat->vfast is normally NO, but may have been set to YES if the -v option
	is used.  If p->vfast is YES, lines NOT matching the pattern will be found.
	Also, if pat->vfast is YES, the counts pat->N_start and pat->N_stop
	apply to the lines NOT matching the pattern.

	Any matches found are stored in "pat->matches", along with those
	saved from the previous buffer (if any).

	The total number of matches stored in "pat->matches" (including those
	saved from the last pass) is stored in "pat->nmatches".  This value is
	also returned by this routine.

Global Variables:

	None

Files:
	None

Return Status:

	Number of matches found, including those saved from the previous pass

*******************************************************************************/

int
pattern_match(pat, begin, end, nsaved)
register PATTERN *pat;
CHAR	*begin, *end;
int	nsaved;
{
	pat->done = NO;
	/* pat->loc is used to prevent matching the same text twice
	 * when multiline matching.
	 */
	pat->loc.loc1 = end+1;
	pat->loc.loc2 = begin-1;

	/* skip over any saved matches from last buffer */
	pat->nmatches = nsaved;

	/* Leave an empty slot when inverting matches (-v) */
	if (pat->vfast) {
		pat->nmatches++;
	}

	/* Do the matching */
	(void)(*pat->procfn)(pat, pat->succfn, begin, end);

	/* If inverting matches, do it. */
	if (pat->vfast) {
		pat->nmatches = vfast_reverse(pat, begin, end, nsaved);
	}
	return(pat->nmatches);
} /* end of function pattern_match() */

/******************************************************************************

Routine:
	vfast_reverse()

Syntax:

	static int vfast_reverse(pat, begin, end, nsaved)

	PATTERN *pat;	// Pointer to pattern to match
	CHAR	*begin;	// Pointer to first character of text to be searched
	CHAR	*end;	// Pointer to last character of text to be searched
	int	nsaved;	// Number of stored matches left over from the previous
			//	text buffer that must be saved.  Note that these
			//	matches must all be BEFORE "begin".  None can be
			//	within the "begin" to "end" range.
Description:

	"cgrep" has three different implementations of the -v and -V options,
	depending on compatibility with other options, and on optimizing
	performance.  One of the implementations is implemented by this
	routine.  The use of this particular implementation was indicated
	by previously setting "pat->vfast".

	In this implementation, pattern_match() calls vfast_reverse() after
	all the matches to the pattern in the text buffer have been recorded.
	vfast_reverse() "inverts" the matches, that is, a line is recorded as
	matched if and only if it was NOT matched in the search.

Global Variables:

	None

Files:
	None

Return Status:

	Number of "matches" found, including those saved from the previous pass

*******************************************************************************/

static int
vfast_reverse(pat, begin, end, nsaved)
register PATTERN *pat;
CHAR	*begin, *end;
int	nsaved;
{
	register CGWINDOW *to, *from;

	/* Set pointer to empty slot. */
	to = pat->matches + nsaved;

	/* Set pointer to first newly found match. */
	from = to + 1;

	/* Now invert the matches.  Adjacent areas of text outside the
	 * matches are stored as matches in place.  There is no need
	 * to "delete" the uninverted matches, as they are simply written
	 * over as needed after being used.
	 */
	if (pat->nmatches == from - pat->matches) {
		/* No matches were found.
		 * Store the entire text as one inverted match.
		 */
		to += store_it(pat, begin, end, to);
	} else {
		/* Store any unmatched text before the first match. */
		to += store_it(pat, begin, from->begin, to);

		/* Loop over the matches, storing the unmatched spaces between
		 * them.
		 */
		for (; from < pat->matches + pat->nmatches - 1; from++) {
			to += store_it(pat, from->end, (from+1)->begin, to);
		}
		/* Store any unmatched text after the last match. */
		to += store_it(pat, from->end, end, to);
	}
	/* Set the number of inverted matches, including those saved from
	 * previous pass.
	 */
	pat->nmatches = to - pat->matches;
	return(pat->nmatches);
} /* end of function vfast_reverse() */

/******************************************************************************

Routine:
	store_it()

Syntax:

	static int store_it(pat, begin, end, to)

	PATTERN *pat;	// Pointer to pattern to match
	CHAR	*begin;	// Pointer to first character of text to be stored
	CHAR	*end;	// Pointer to last character of text to be stored
	CGWINDOW *to;	// Pointer to array element in pat->matches
			//	in which to store the match.

Description:

	Called from vfast_reverse() to possibly store a group of contiguous
	lines of unmatched text as a single inverted match.

	Note: Each unmatched line of text counts as one match for
	"cgrep -v -N", OR "cgrep -v +N" purposes, but multiple adjacent lines
	are stored as a single match in pat->matches, and counted as a
	single match in pat->nmatches.

Global Variables:

	None

Files:
	None

Return Status:

	1	if a match was stored
	0	if no match was stored

*******************************************************************************/

static int
store_it(pat, begin, end, to)
register PATTERN *pat;
CHAR	*begin, *end;
register CGWINDOW *to;
{
	CHAR *p = begin;

	/* Don't store zero length text, or if maximum match count has been
	 * reached previously.
	 */
	if (begin >= end || pat->done) {
		return(0);
	}
	/* If (inverted) matches are being skipped, or if there is an upper
	 * limit on the number of matches that may be stored, we need to
	 * account for this.
	 */
	while (pat->N_start || pat->N_stop) {
		if (check_count(pat)) {
			/* 
			 * The match is not to be skipped.  But since there
			 * is an upper limit on the number of matches, we
			 * must advance one line at a time.
			 * (check_count() maintains the count, and has set
			 * pat->done to YES if the limit was just reached).
			 */
			p = memchr_fwd(p+1, NL);

			/* See if we are done. */
			if (p >= end || pat->done) {
				/* We've reached the end of the text or the maximum
				 * number of matches (i.e. lines).  Set "end" to
				 * delimit the text to store.  Then, break out of
				 * the loop so that the entire set of lines
				 * processed so far will be stored as a match.
				 */
				end = MIN(p, end);
				break;
			}
		} else {
			/* 
			 * A match is to be skipped.  Therefore, we pass over one
			 * line. (check_count() maintains the count.)
			 */
			p = begin = memchr_fwd(begin+1, NL);
			/* If there is no remaining text,
			 * exit without storing a match.
			 */
			if (begin >= end) {
				return(0);
			}
		}
	}
	/* Store match */
	to->begin = begin;
	to->end = end;
	to->wbegin = to->wend = 0;
	to->prev_matches = 0;
	/* Return the fact that one match was stored */
	return(1);
}

/******************************************************************************

Routine:
	re_lnexec()

Syntax:

	BOOL re_lnexec(pat, fcn, s, e)

	PATTERN *pat;	// Pointer to pattern to match
	BOOL (*fcn)();	// If non-null, a pointer to the successor matching function.
			//	In this case, text matched by re_lnexec() must also
			//	match in every successor function to be considered
			//	a full match.
	CHAR	*s;	// Pointer to first character of text to be searched
	CHAR	*e;	// Pointer to last character of text to be searched

Description:

	One of the pattern matching functions used by pattern_match().  Used
	when every line of text needs to be matched (matching in the forward
	direction).  If "fcn" is non-zero, the text must also match that
	function and its successors.

Global Variables:

	CHAR *Mline_mark;
	CHAR *F1;
	CHAR *S1;

	All three of these global variables are used here only to avoid spurious
	matches if both of the following hold:

	1) pat->vflag has been set (i.e. we are searching line by line for lines
		that do NOT match)
	   AND
	2) the pattern to be matched can span multiple lines.

	See the comments where these variables are used.
	

Files:
	None

Return Status:

	YES	At least one match was found, and the search was terminated.

	NO	The entire text was searched for matches, regardless
		of whether or not any were found.


*******************************************************************************/

BOOL 
re_lnexec(pat, fcn, s, e)
PATTERN	*pat;
BOOL (*fcn)();
register CHAR *s, *e;
{
	CHAR *nb, *ne;
	register CHAR *save_s, *max_s;
	BOOL ret;

	save_s = s;
	/* pat->vflag may sometimes be set if the -v option was used. */ 
	/* If vflag is set, don't accept any "matches" after Mline_mark,
	 * because they could be negated by matches in the next buffer.
	 * We can't reset "e" to Mline_mark, because multiline searches can
	 * go outside the line, past Mline_mark.
	 */
	max_s = ((e - Mline_mark > 0 && pat->vflag) ? Mline_mark : e);
	/* Loop over lines of text. */
	while (s < max_s) {
		/* In the multiline case, grepmatch will extend its search
		 * outside the line!
		 */
		/* Initialize the text line delimiters to NULL. grepmatch()
		 * needs these.
		 */
		S1 = F1 = 0;
		/* We have a partial match on every line.  Call grepmatch()
		 * to check for and store a possible full match on this line,
		 * (subject to the count constraints of pat->N_start and
		 * pat->N_stop).
		 */
		ret = grepmatch(pat, fcn, save_s, e, s, s, &nb, &ne);
		if (ret && !pat->vflag) {
			/* We have reached the maximum number of matches.
			 * Exit, since we no longer look for matches.
			 */
			return(YES);
		}
		if (!ret && pat->vflag) {
			/* We have no full match, but we are using -v.  Therefore,
			 * treat it as a potential match.  In this case, fixloc()
			 * was not called by grepmatch(), and needs to be called
			 * here to check the counts (N_start, N_stop), and,
			 * if needed, delimit and store the match.
			 */
			ret = fixloc(pat, nb, ne);
			if (ret) {
				/* Match stored. Exit. */
				return(YES);
			}
			/* Match skipped, or match count exceeded.  Keep going. */
		}
		/* Advance "s" to next line. */
		s = ne;
	}
	/* No full match found. */
	return(NO);
}

/******************************************************************************

Routine:
	re_lbexec()

Syntax:

	BOOL re_lbexec(pat, fcn, s, e)

	PATTERN *pat;	// Pointer to pattern to match
	BOOL (*fcn)();	// If non-null, a pointer to the successor matching function.
			//	In this case, text matched by re_lnexec() must also
			//	match in every successor function to be considered
			//	a full match.
	CHAR	*s;	// Pointer to first character of text to be searched
	CHAR	*e;	// Pointer to last character of text to be searched

Description:

	One of the pattern matching functions used by pattern_match().  Used
	when every line of text needs to be matched (matching in the backward
	direction).  If "fcn" is non-zero, the text must also match that
	function and its successors.

Global Variables:

	CHAR *F1;
	CHAR *S1;

	Both of these global variables are used here only to avoid spurious
	matches if both of the following hold:

	1) pat->vflag has been set (i.e. we are searching line by line for lines
		that do NOT match)
	   AND
	2) the pattern to be matched can span multiple lines.

	See the comments where these variables are used.
	

Files:
	None

Return Status:

	YES	At least one match was found, and the search was terminated.

	NO	The entire text was searched for matches, regardless
		of whether or not any were found.


*******************************************************************************/

BOOL 
re_lbexec(pat, fcn, s, e)
PATTERN	*pat;
BOOL (*fcn)();
register CHAR *s, *e;
{
	CHAR *nb, *ne;
	register CHAR *save_e;
	BOOL ret;

	save_e = e;
	/* Loop over lines of text, backwards. */
	while (e > s) {
		/* In the multiline case, grepmatch will extend its search
		 * outside the line!
		 */
		/* Initialize the text line delimiters to NULL. grepmatch()
		 * needs these.
		 */
		S1 = F1 = 0;
		/* We have a partial match on every line.  Call grepmatch()
		 * to check for and store a possible full match on this line,
		 * (subject to the count constraints of pat->N_start and
		 * pat->N_stop).
		 */
		ret = grepmatch(pat, fcn, s, save_e, e, e, &nb, &ne);
		if (ret && !pat->vflag) {
			/* We have reached the maximum number of matches.
			 * Exit, since we no longer look for matches.
			 */
			return(YES);
		}
		if (!ret && pat->vflag) {
			/* We have no full match, but we are using -v.  Therefore,
			 * treat it as a potential match.  In this case, fixloc()
			 * was not called by grepmatch(), and needs to be called
			 * here to check the counts (N_start, N_stop), and,
			 * if needed, delimit and store the match.
			 */
			ret = fixloc(pat, nb, ne);
			if (ret) {
				/* Match stored. Exit. */
				return(YES);
			}
			/* Match skipped, or match count exceeded.  Keep going. */
		}
		/* Back up "e" to previous line. */
		e = nb;
	}
	return(NO);
}

/******************************************************************************

Routine:
	re_nullexec()

Syntax:

	BOOL re_nullexec()

Description:

	One of the pattern matching functions.  Used when no line of text
	is to be matched.

Global Variables:

	None

Files:
	None

Return Status:

	NO

*******************************************************************************/

BOOL 
re_nullexec()
{
	return(NO);
}

/****************************************************************************

Routine:
	fixloc()

Syntax:
	static BOOL fixloc(pat, mb, me)

	PATTERN *pat;	// Pointer to pattern matched
	CHAR	*mb;	// Pointer to first character of (expanded) match
	CHAR	*me;	// Pointer to last character of (expanded) match

Description:

	If pat->vflag is NO, fixloc() is called by grepmatch() in the event
		of a full pattern match.

	If pat->vflag is YES, fixloc() is called by re_lnexec() or re_lbexec()
		if no match is found to a line, in which case, we treat it
		as a match.

	The match has already been expanded to complete lines
		(unless pat->pbounds is YES, in which case it is NOT to be
		expanded to complete lines). 

	fixloc() does the following:

	1) If the pattern to be matched spans multiple lines, it may have
	already been found, since text may be rescanned.  If so, skip it.

	2) Calls check_count() to check pat->N_start, pat->N_stop, and pat->done
	to see if we should ignore this match.
	NOTE: if pat->vfast is YES, check_count() is called elsewhere instead.

	3) Either stores the match, or combines it with previous match as
	appropriate.

	4) Returns NO unless we have reached the upper limit to the number
	of matches (if any), in which case it returns (YES).

Global Variables:

	None

Files:
	None

Return Status:

	YES	if the maximum number of matches has been reached.
	NO	if the maximum number of matches has not been reached,
		or there is no such maximum.

*******************************************************************************/

static BOOL
fixloc(pat, mb, me)
PATTERN *pat;
CHAR *mb, *me;
{
	BOOL join;
	BOOL ret;
	CGWINDOW *prev = 0;

	/* The pattern can span multiple lines of text.  Since in this case
	 * we sometimes rescan, we may have already found it.  The limits
	 * of the previous scans are stored in pat->loc.loc1 and pat->loc.loc2.
	 */
	if (pat->mult_lines) {
		/* If already passed this point, skip this match. */
		if (!pat->backward && me < pat->loc.loc2) {
			return(NO);
		}
		/* If already at this point, skip this match unless it
		 * extends before the previous limits.
		 */
		if (!pat->backward && me == pat->loc.loc2 && mb >= pat->loc.loc1 ) {
			return(NO);
		}
		/* If already passed this point (going backward), skip this match. */
		if (pat->backward && mb > pat->loc.loc1) {
			return(NO);
		}
		/* If already at this point (going backward), skip this match unless
		 * it extends after the previous limits.
		 */
		if (pat->backward && mb == pat->loc.loc1 && me <= pat->loc.loc2) {
			return(NO);
		}
		pat->loc.loc1 = mb;
		pat->loc.loc2 = me;
	}
	/* Check if we must skip this match, or if the maximum number of
	 * matches has been reached.  (If pat->vfast is set, this check
	 * must be done on inverted matches, so we cannot do it here).
	 */
	ret = (BOOL)(pat->vfast || check_count(pat));
	if (!ret) {
		/* We skip this match. */
		return(pat->done);
	}

	/* If we are matching a delimiter, just store the match in the one
	 * slot allocated, and exit.  Since we must count these matches, we do
	 * not join them.
	 * To save memory, we have allocated space for only one match, so
	 * only the last match will be saved (the previous one is overwritten).
	 */
	if (pat->is_delim) {
		pat->nmatches++;
		pat->matches[0].begin = mb;
		pat->matches[0].end = me;
		return(pat->done);
	}

	/* We must determine if this match should be joined with previous
	 * match.  Get the previous match, if any.
	 */
	if (pat->nmatches) {
		prev = &pat->matches[pat->nmatches - 1];
	}

	if (!pat->nmatches || mb > prev->end || me < prev->begin) {
		/* No previous stored match, or matches are disjoint */
		join = NO;
	} else if (pat->pbounds &&
	    !(mb >= prev->begin && me <= prev->end) &&
	    !(mb <= prev->begin && me >= prev->end)) {
		/* pbounds is set, and neither match completely contains the other.
		 * (If pbounds is set, we do not join adjacent or overlapping
		 * matches, but we still join matches if one completely contains
		 * the other.)
		 */
		join = NO;
	} else {
		/* Match touches previous one.  Join them into a single match,
		 * as this saves memory and speeds up performance.
		 */
		join = YES;
	}
	/*  Join or store the match. */
	if (join) {
		/* If we extend the beginning (end) of an existing match,
		 * we null out the window limit, just in case the match has
		 * already been windowed.  It will then be windowed again.
		 */
		if (mb < prev->begin) {
			prev->begin = mb;
			prev->wbegin = 0;
		}
		if (me > prev->end) {
			prev->end = me;
			prev->wend = 0;
		}
	} else {
		/* Store a new match. */
		CGWINDOW *p;

		/* Get more memory if needed. */
		match_alloc(pat);
		p = &pat->matches[pat->nmatches];
		pat->nmatches++;
		p->begin = mb;
		p->end = me;
		p->wbegin = p->wend = 0;
		p->prev_matches = 0;
	}
	return(pat->done);
}

/******************************************************************************

Routine:
	grepmatch()

Syntax:

	BOOL grepmatch(pat, fcn, rs, re, mb, me, pnb, pne)

	PATTERN *pat;	// Pointer to pattern matched
	BOOL (*fcn)();	// If non-null, a pointer to the successor matching function.
			//    In this case, we currently have only a partial match.
			//    The text must match the pattern in every successor
			//    function to be considered a full match.
			//    If NULL, there is a full match.
	CHAR	*rs;	// Pointer to first character of text being scanned
	CHAR	*re;	// Pointer to last character of text being scanned
	CHAR	*mb;	// Pointer to first character of match
	CHAR	*me;	// Pointer to last character of match
	CHAR	**pnb;	// Returns pointer to first character
			//    of possibly extended match
	CHAR	**pne;	// Returns pointer to last character
			//    of possibly extended match

Description:

	grepmatch() is called if we have a full or partial match to a pattern.

	grepmatch() does the following:

	1) extends the match to fill lines (Exception: if a full match,
	and pat->pbounds is TRUE, there is no extension).

	2) In the full match case, ("fcn" is NULL) calls fixloc() to store
	the extended match if it should be stored, and passes back the
	extended match limits in "pnb" and "pne".  The return code from
	fixloc() (usually NO) is returned.

	Exception: if matching a pattern that can span multiple lines,
	passes back the unextended match, event though the extended match
	may have been stored.  This is so overlapping multiline matches will
	not be missed.

	Exception: if pat->vflag is true, fixloc() is not called, and YES
 	is returned.

	3) In the partial match case, calls "fcn" to do a fuller match, and passes
 	back the extended partial match.  (If multiline matching, the range of
 	text for the fuller match is increased to include sufficient
 	surrounding lines).  The return code from "fcn" is returned.

Global Variables:

	CHAR *F1;
	CHAR *S1;

	Both of these global variables are used here only to avoid spurious
	matches if both of the following hold:

	1) pat->vflag has been set (i.e. we are searching line by line for lines
		that do NOT match)
	   AND
	2) the pattern to be matched can span multiple lines.

	See the comments where these variables are used.

Files:
	None

Return Status:

	if pat->vflag == NO, the return code will be:

	YES	The maximum number of full matches has been reached.
			No more matching is needed.
	NO	The maximum number of full matches has not been reached.

	if pat->vflag == YES the return code will be:

	YES	We have a full match to the pattern.
	NO	We do not have a full match to the pattern.

	In most instances, the return code is not set in grepmatch() itself.
	Rather, the return code from fixloc() or (*fcn)() is passed back.

*******************************************************************************/

BOOL
grepmatch(pat, fcn, rs, re, mb, me, pnb, pne)
PATTERN *pat;
BOOL (*fcn)();
CHAR *rs, *re;
CHAR *mb, *me, **pnb, **pne;
{
	register CHAR *s, *f;
	register CHAR *s1, *f1;
	register int i;
	BOOL (*nextfcn)();
	BOOL ret;

	/* If matching a pattern that may span multiple lines, and we have
	 * a supposed full match, and vflag is on, we must make sure that the
	 * full match partially overlaps the original line.  This check is
	 * needed because we have extended the original line to contain some
	 * surrounding lines.
	 */
	if (!fcn && pat->mult_lines && pat->vflag &&
	    S1 && F1 && (me <= S1 || mb >= F1)) {
		/* The supposed full match does not overlap the line that
		 * was originally selected.  The match is spurious.
		 * Therefore, we return NO, to indicate no full match.
		 */
		*pnb = mb;
		*pne = me;
		return(NO);
	}
	/* If we have a full match, and pat->pbounds is set, don't store
	 * null matches and don't extend the match to complete lines.
	 */ 
	if (!fcn && pat->pbounds) {
		*pnb = mb;
		*pne = me;
		/* Don't store null match */
		if (me < mb) return(pat->vflag);
		/* Store unextended match and return.  We don't store
		 * if pat->vflag is set.
		 */
		return((BOOL)(pat->vflag || fixloc(pat, mb, me)));
	}

	/* We now expand the match to full lines of text, unless we
	 * have already done so.
	 */
	if (fcn != pat->succfn && !pat->mult_lines) {
		/* This is not the first stage of the match, and pattern can
		 * match only a single line.  In this case, we can reuse the
		 * line delimiters from the previous stage of the match,
		 * since they must be the same.
		 */
		s = rs;
		f = re;
	} else {
		/* Either this is the first stage of the match, in which case
		 * we have not found the line delimiters before, or the
		 * pattern can match multiple lines, in which case this match
		 * may occupy fewer lines than a possible earlier stage.
		 * We must find the line delimiters.
		 */
		/* Set the delimiters to the end and beginning of the match. */
		f = me;
		s = mb;

		/* A match to the null string needs special handling. */
		if (f < s) {
			/* Adjustment for null match.  Expand it to match
			 * the next single character.
			 */
			if (pat->backward) {
				s = f;
			} else {
				f = s;
			}
		}
		/* A match to a single new-line character needs special handling. */
		if (f == s && *s == NL) {
			/* Adjustment for match to single new-line.
			 * Expand it to also include the next character.
			 */
			if (pat->backward) {
				--s;
			} else {
				f++;
			}
			/* If we matched the new-line character delimiting the
			 * end of our search range, ignore the match.
			 */
			if (f > re || s < rs) {
				*pnb = mb;
				*pne = me;
				return(NO);
			}
		}
		/* Now set the line delimiters by searching for the
		 * "next" new-line characters.
		 */
		s1 = s = memchr_back(s, NL);
		f1 = f = memchr_fwd(f, NL);
		/* The "global" delimiters S1 and F1 save the line delimiters
		 * from the first stage of the match.  Don't reset them if they
		 * have been set before.
		 */
		S1 = (S1 ? S1 : s1);
		F1 = (F1 ? F1 : f1);
	}
		
	/* Is this a full match? */
	if (!fcn) {
		/* Set the match delimiters to return to the caller.  If the
		 * pattern must be on a single line, we send back the line
		 * delimiters, since there is no need to search the rest of
		 * the line for more matches.  Otherwise, we pass back the
		 * unextended match limits, since we have to keep searching
		 * from the last matched character, or we may miss overlapping
		 * matches.
		 */
		*pnb = (pat->mult_lines ? mb : s);
		*pne = (pat->mult_lines ? me : f);
		/* Store extended match and return.  We don't store
		 * if pat->vflag is set.
		 */
		mb = s;
		me = f;
		return((BOOL)(pat->vflag || fixloc(pat, mb, me)));
	}

	/* We have only a partial match at this stage */
	/* If the full pattern can match multiple lines, we extend the partial
	 * match to enough surrounding lines so that we don't miss a full
	 * match in a later stage.
	 * NOTE:  This extension is used only to expand the range of
	 * subsequent stages of the match.  It is not passed back to the
	 * caller.
	 */
	for (i = 0; i < pat->mult_lines; i++) {
		if (f < re && pat->procfn != re_lnexec) {
			f = memchr_fwd(f+1, NL);
		}
		if (s > rs && pat->procfn != re_lbexec) {
			s = memchr_back(s-1, NL);
		}
	}
	/* Ensure that any extension did not go outside the search range.
	 */
	if (f > re) {
		f = re;
	}
	if (s < rs) {
		s = rs;
	}
	/* Get a pointer to the matching function for the next stage of the match. */
	nextfcn = (fcn == pat->procfn ? pat->succfn :
		  (fcn == pat->succfn ? pat->thirdfn : 0));
	/* Perform the subsequent stages of the match on the range of the
	 * fully extended partial match.
	 */
	ret = (*fcn)(pat, nextfcn, s, f);
	/* Pass back the delimiters of the line extended match.  There is no
	 * need to check the rest of the current line for partial matches,
	 * (even if the pattern can span multiple lines) because the entire
	 * line has been checked for full matches.
	 */
	*pnb = s1;
	*pne = f1;

	/* Return the return code from the subsequent stages of the match. */
	return(ret);
}

/******************************************************************************

Routine:
	check_count()

Syntax:

	BOOL check_count(pat)

	PATTERN *pat;	// Pointer to pattern that was matched

Description:

	check_count() is called in the event of a full match to a pattern
	(in the -v case, in the event a line does not match the pattern)
	It determines if the match is to be kept, or skipped over.

	check_count() does the following:

	1) If pat->done is set:
	Then the upper limit on the number of matches to this pattern
	has already been exceeded.  check_count() does nothing, and returns NO.

	2) else if pat->N_start is non-zero:
	Then the next N_start matches are to be skipped.  check_count()
	decrements N_start, and returns NO.

	3) else if pat->N_stop is non-zero:
	Then there is an upper limit on the number of matches to this pattern,
	but it has not yet been exceeded.  check_count() decrements N_stop,
	sets pat->done if N_stop has reached zero, and returns YES.

	4) else:
	Then matches are not being counted. check_count() returns YES.

Global Variables:

	None

Files:
	None

Return Status:

	YES	if the match is to be saved.
	NO	if the match is to be skipped.

*******************************************************************************/

static BOOL
check_count(pat)
PATTERN *pat;
{
	/* If maximum number of matches already reached, don't keep the match.  */
	if (pat->done) {
		return(NO);
	}

	/* If skipping matches, decrement skip count and don't keep the match.  */
	if (pat->N_start) {
		--pat->N_start;
		return(NO);
	}
	/* If there is a maximum number of matches, decrement the count, and,
	 * if it reaches zero, set pat->done to stop future matches.
	 * However, we keep this match.
	 */
	if (pat->N_stop) {
		--pat->N_stop;
		if (!pat->N_stop) {
			pat->done = YES;
		}
	}
	/* If not counting matches, and not skipping any, keep match. */
	return(YES);
}

#ifdef __MSDOS__
#define		MATCHSTEP	3640
#define		MATCHLIMIT	MATCHSTEP
#else
#define		MATCHSTEP	8192
#define		MATCHLIMIT	(64*8192)
#endif

/******************************************************************************

Routine:
	match_alloc()

Syntax:

	void match_alloc(pat)

	PATTERN *pat;	// Pointer to pattern to match

Description:

	match_alloc() is called before a pattern match is to be stored
	to make sure that there are enough memory slots to store the match.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/


void
match_alloc(pat)
register PATTERN *pat;
{
	if (pat->match_buffers == 0) {
		/* There are no slots allocated.  We must allocate some. */
		pat->match_buffers = MATCHSTEP;
		pat->matches = (CGWINDOW *)cgmalloc2(pat->match_buffers,
			sizeof(CGWINDOW), "pattern match buffers");
	} else if (pat->nmatches >= pat->match_buffers) {
#ifdef	DEBUG
		printf("OLD MATCH BUFFERS %d\n", pat->match_buffers);
#endif
		/* There is no available slot.  We must allocate more.
		 * If we are not over MATCHLIMIT slots, we double
		 * the number of slots.  Otherwise, we just add
		 * MATCHSTEP more slots.
		 */
		if (pat->match_buffers < MATCHLIMIT) {
			pat->match_buffers *= 2;
		} else {
			pat->match_buffers += MATCHSTEP;
		}
#ifdef	DEBUG
		printf("NEW MATCH BUFFERS %d\n", pat->match_buffers);
#endif
		/* We do a realloc to save the previous contents. */
		pat->matches = (CGWINDOW *)cgrealloc2((char *)pat->matches,
			pat->match_buffers, sizeof(CGWINDOW),
			"pattern match buffers");
	}
}
