#ifndef _RE_DEFINED
#define _RE_DEFINED	1

#ifdef DEBUG
#define		TRACE(n)	(n < debug)
extern int debug;
#define		EPRINTSIZE	50000
#endif

typedef unsigned char CHAR;

typedef enum BOOL {
	NO = 0,
	YES = 1
} BOOL;

/* Number of possible text characters (8-bit ASCII) */
#define NCHARS	256

/* Modes of parsing. */
typedef enum PARSE_TYPE {
	grepparse,	/* "grep" mode.  Will be converted to one of the others. */
	egrepparse,	/* "egrep" mode. */ 
	fgrepparse,	/* "fgrep" mode. */
	nlparse,	/* Pattern matches every line.  No parsing needed. */
	nullparse	/* Pattern cannot match anything.  No parsing needed. */
} PARSE_TYPE;

/*************** Structures for Commentz-Walter algorithm **************************/

/* All Commentz-Walter structure members are initialized before or during
 * compilation, and are used, but not modified, by the matching algorithm.
 */

/* NODE:  Auxiliary Structure used for Commentz-Walter algorithm. */
typedef struct NODE {
	short matched;		/* Is this state a match? */
	short depth;		/* Distance of <llit> from end of string.
				 * e.g. 1 means last character of string.
				 * 0 means this is the ROOT node.
				 */
	short shift1;
	short shift2;
	long id;		/* Unique identifier for this NODE.
				 * Used only for debugging.
				 */
	struct NODE *child;	/* Pointer to first NODE structure in chain
				 * corresponding to characters that can
				 * precede this NODE in a match.  Every node
				 * in this chain will have <depth> one greater
				 * than the <depth> of this NODE.
				 */
	struct NODE *sibling;	/* Pointer to sibling NODE structure in chain.
				 * This chain is the "followset" of a NODE
				 * of one less depth.  Any node in this
				 * chain will have the same <depth> as this
				 * NODE.
				 */
	struct NODE *lips;	/* Pointer to NODE that corresponds to the
				 * Longest Initial Proper Substring (LIPS) of
				 * this node.
				 * For example, if the pattern is "abcd|xab",
				 * the NODE for "bcd" will have <lips>
				 * pointing to the NODE for "b" (there is no
				 * "bc" NODE).
				 * Used only in compilation to calculate
				 * <shift1> and <shift2>.
				 * <lips> will always have a smaller <depth>
				 * than this NODE or will be the root NODE.
				 */
	struct NODE *link;	/* Temporary pointer used to make a linked list
				 * of NODEs.
				 */
	CHAR llit;		/* Character to be matched. */
} NODE;

/* RE_CW:  Structure used for Commentz-Walter algorithm. */
typedef struct RE_CW {
	unsigned mindepth;	/* Minimum length of strings to be matched. */
	unsigned step[NCHARS];	/* Minimum step size after partial match. */
	unsigned delta0[NCHARS];/* Same as <step>, with LARGE replacing 0. */
	CHAR *cmap;		/* Character map. */
	NODE *root;		/* Pointer to root NODE. */
	BOOL fullextent;	/* Do we need to know the full extent of the match? */ 
	struct NODE **tab;	/* Used for root NODE only to avoid traversing
				 * long chain.  NODE transition table based
				 * on next text character.
				 */
	unsigned npat;		/* Number of patterns to be matched. */
	long nodeid;		/* Count of NODES. */
	CHAR *bmpat;		/* Single pattern to be matched. */
	int *shift1;		/* Step size after single pattern partial match. */
} RE_CW;

/*************** Enumeration for Regular Expression matching ***********************/

/* Expression types used in compiled regular expression.
 * Also, token types used in parsing regular expressions.
 */
typedef enum EXPR_TYPE {
	/* "lit" expression types */

	EOP = 0,	/* End of pattern.  Token is '\0' at end of expression  */
	Literal,	/* Matches a specific character. */
			/* Token is any character not used for another
			 * purpose, or any character escaped by a backslash.
			 */
	Dot,		/* Matches any character except a new-line. */
			/* Token is '.'. */
	Charclass,	/* Matches any one of a set of characters. */
			/* Token is '[', (terminated by ']'. */

	/* non-lit expression types */

	Backref,	/* Used in back-referencing.
			 * Matches identical text that last matched
			 * corresponding Group. Internal token is '`'
			 * (previously converted from \<digit>).
			 */
	Group,		/* Identifies group for backreferencing.
			 * Internal token is '(', terminated by ')',
			 * (previously converted from "\(" and "\)").
			 * NOTE: "()" tokens do not generate a Group
			 * unless backreferencing is used.
			 */
	Cat,		/* Matches concatenation of left and right subexpressions.
			 * There is no token.  Adjacent "lit" expressions are
			 * automatically concatenated.
			 */
	Alternate,	/* Matches either left or right subexpression.
			 * Token is <new-line>.
			 */
	Quest,		/* Matches zero or one match to subexpression.
			 * Token is '?'.
			 */
	Star,		/* Matches zero or more matches to subexpression.
			 * Token is '*'.
			 */
	Plus,		/* Matches one or more matches to subexpression.
			 * Token is '+'.
			 */

	/* The following are not used as expression types, but are used
	 * only as tokens in parsing.
	 */

	Non_alpha,	/* Gets converted to Charclass matching word delimiter
			 * (any character other than letter, number or underscore).
			 * Internal token is '#', either converted from
			 * "\<", or "\>", or inserted because of -x1 option.
			 */
	Lpar,		/* Delimits the start of an explicit  subexpression.
			 * Token is '('.
			 */
	Rpar,		/* Delimits the end of an explicit subexpression.
			 * Token is ')'.
			 */
	NUM_EXPR_TYPES
} EXPR_TYPE;

/*************** Structures  for Regular Expression matching ***********************/

/* Unique ID for each "lit" expression. */
typedef int ID;

/* FID:  Auxiliary Structure used in matching a Regular Expression.
 * A FID corresponds to a "lit" expression, with some auxiliary
 * information needed when doing approximate matching.
 * FID's are used in a STATE structure.
 */
typedef struct FID {
	ID	id;	/* Lit Expression id */
	short	cost;	/* Accumulated mismatch cost (when approx. matching). */
	unsigned short ins_cost; /* Incremental mismatch cost of one insertion
			 * before this "lit", while in the STATE.
			 */
} FID;

/* FFID:  Auxiliary Structure used in matching a Regular Expression.
 * A FFID corresponds to a "lit" expression, with some auxiliary
 * information needed when doing approximate matching.
 * FFID's are used in a FOLLOW_SET structure.
 * Although FFIDs and FIDs could share the same structure template, they
 * are not functionally identical.
 */
typedef struct FFID {
	ID	id;	/* Lit Expression id */
	short	cost;	/* Incremental mismatch cost due to all deletions between
			 * this "lit" and the "lit" to which the FOLLOW_SET
			 * belongs.
			 */
	unsigned short ins_cost; /* Incremental mismatch cost of one insertion
			 * between this "lit" and the "lit" to which the FOLLOW_SET
			 * belongs.
			 */
} FFID;

/* FOLLOW_SET:  Auxiliary Structure used in matching a Regular Expression.
 * Each EXPR that is a "lit" expression contains two FOLLOW_SETs,
 * one containing the set of "lit" expressions that may immediately follow
 * the given "lit" expression in a text match in the forward direction, and
 * the other the same for matches in the backward direction.  Each array
 * of FFIDs dynamically filled in as needed during text matching.
 */
typedef struct FOLLOW_SET {
	unsigned flen;		/* Number of lit expressions */
	FFID *follow;		/* Array of FFIDs of following lit expressions */
} FOLLOW_SET;

/* EXPR:  The atomic building block of a Regular Expression. */
typedef struct EXPR {

	/* This first group of members are initialized during compilation
	 * of the regular expression, and are used, but not modified,
	 * during the text matching.
	 */

	ID id;			/* Unique ID of "lit" expression  */
	FOLLOW_SET b;		/* Following "lit" expressions for backward match */
	FOLLOW_SET f;		/* Following "lit" expressions for forward match */
	struct EXPR *l;		/* Pointer to Left child */
        struct EXPR *r;		/* Pointer to Right child (or CCL mask) */
        struct EXPR *parent;	/* Pointer to Parent */
	CHAR lit;		/* Literal character or tag */
	unsigned short n_bracket;/* Disallow mismatch in approx. match (bracket #) */
	EXPR_TYPE type:5;	/* Type of expression (in grammar) */

	/* The following members are used temporarily during the optimization
	 * steps of compilation.  They are not used by the matching algorithm.
	 */
	unsigned dot:1;		/* Has Dot and no non-nullable non-Dots. */
	unsigned dot_star_l:1;	/* Has Dot Star or Dot Plus on left. */
	unsigned dot_star_r:1;	/* Has Dot Star or Dot Plus on right. */
	unsigned rept:1;	/* +, *, ?, here or below */
	unsigned alter:1;	/* Interior alternation below */
	unsigned backref_no_star:1;/* Backref here or below, excl. brstar */

	/* The following members are used during compilation.  They are not
	 * used directly by the matching algorithm.
	 *
	 * "min_length_ex" is used temporarily during the optimization steps
	 * of the compilation.
	 *
	 * "group" and "backref", "min_length", and "max_length_ex" are used to
	 * determine the values of "backref", "min_length", and "max_length",
	 * respectively, in the RE_RE structure for the regular expression as
	 * a whole.
	 *
	 * "nl_first", "nl_last", "nl_alt", and "nl_count" are used to determine
	 * "multlines" in the PATTERN structure for the pattern as a whole.
	 */
	unsigned group:1;	/* Backref group here or below */
	unsigned backref:1;	/* Backref here or below */
	unsigned nl_first:1;	/* NL (or CCL with NL) is first lit in dominant. */
	unsigned nl_last:1;	/* NL (or CCL with NL) is last lit in dominant. */
	unsigned nl_alt:1;	/* New-Line Ambiguity due to Interior Alternate */
	short nl_count; 	/* New-Line Count */
	short min_length_ex;	/* Minimum length of exact matching text */
	short min_length;	/* Minimum length of matching text */
	short max_length_ex;	/* Maximum length of exact matching text */
} EXPR;

/* STATE:  Auxiliary Structure used in matching a Regular Expression.
 * Each STATE structure represents a state of the Finite-State-Machine
 * algorithm for regular expression matching.  STATE structures are
 * dynamically created during text matching as each state is entered
 * for the first time, except that the initial state is created just
 * before text matching.
 */
typedef struct STATE {
	struct STATE *tab[NCHARS];/* State transition table. Points to the
				 * state that will be entered based on the next text
				 * character, or is 0, if this particular state
				 * transition has not yet been encountered.
				 */
	BOOL is_match;	/* Does this state represents a match to the entire pattern?
			 * This info. is actually redundant, because "is_match"
			 * will be YES if and only if the End of Pattern (EOP)
			 * "lit" is in the POSITION_SET for this state.
			 * However, making this info. more accessible enhances
			 * performance and simplifies code.
			 */
	unsigned npos;	/* Number of FIDs in the POSITION_SET for this state.
			 */
	int pos;	/* Index into "posbase[]" of the first FID structure
			 * comprising the POSITION_SET of the state.
			 */
} STATE;

/* PSFID:  Auxiliary Structure used in matching a Regular Expression.
 * This is a variant of the FID and FFID structures, for use in POSITION_SETs.
 *
 * A PSFID struct is different from a FID or FFID in the following way:
 *
 * The "id" of the "lit" expression is not stored in a PSFID.  Rather,
 * the POSITION_SET contains an array of PSFID's, and the array element
 * with index "id" contains the information about the "lit" expression.
 *
 * If base[id].previd == 0: lit with ID id is NOT in POSITION_SET.
 *
 * If base[id].previd == -1: lit with ID id is first lit in POSITION_SET.
 *
 * If base[id].previd == n (>0): lit with ID id is in POSITION_SET,
 *    and lit with ID n is the previous lit in POSITION_SET.
 */
typedef struct PSFID {
	ID	previd;	/* Lit Expression id of previous FID */
	short	cost;	/* Mismatch cost (when approx. matching). */
	unsigned short ins_cost; /* Cost of insert before. */
} PSFID;

/* POSITION_SET:  Auxiliary Structure used in matching a Regular Expression.
 * These structures can be created and modified during text matching.
 *
 * POSITION_SETs are used only temporarily, in setting up and populating a
 * STATE or a FOLLOW_SET.
 *
 * When used for populating a STATE, once the corresponding STATE is created,
 * the array of FID's has been stored in the "posbase" member of the regular
 * expression structure, and the POSITION_SETs can be reused. The PSFID's in
 * the POSITION_SET are converted to FID's for storage in the STATE structure.
 *
 * Each STATE of the finite state machine is completely determined
 * by the set of FID's that are in the corresponding POSITION_SET
 * for that state.  Each FID corresponds to one "lit" expression.
 * Roughly speaking, a "lit" is in the POSITION_SET for a state if a match
 * to the "lit" would cause a transition to a state that is closer to a match.
 * For example, if the regular expression being matched is "abacd*e | fghij", the
 * POSITION_SET for the initial state consists of the first 'a' and the 'f'.  If
 * the first text character encountered is an 'a', we transition to a state with
 * only the 'b' in its POSITION_SET.
 * A STATE is a match to the pattern if and only if the EOP (End of Pattern)
 * "lit" is in its POSITION_SET.
 *
 * When used for populating a FOLLOW_SET, once the corresponding FOLLOW_SET
 * is created, the array of FFID's has been stored in the FOLLOW_SET structure,
 * and the POSITION_SETs can be reused. The PSFID's in the POSITION_SET are
 * converted to FFID's for storage in the FOLLOW_SET structure.
 *
 * Each FOLLOW_SET of a "lit" expression is completely determined by the set of
 * FFID's that are in the corresponding POSITION_SET for that set.  Each FFID
 * corresponds to one "lit" expression.  Each "lit" expression has two
 * FOLLOW_SETS, one consisting of all the "lit" expressions that can follow
 * the given "lit" expression in matching text in the forward direction, and
 * the other for matches in the backward direction.
 */

typedef struct POSITION_SET {
	unsigned count;	/* Number of "lit" exps in POSITION_SET */
	ID last;	/* ID of last "lit" exp in POSITION_SET.
			 * -1 if POSITION_SET is empty.
			 */ 
	PSFID *base;	/* array of MAXID PSFIDS */
} POSITION_SET;

/* RE_RE:  Structure used for matching a regular expression. */
typedef struct RE_RE {
	/* This first group of attributes are determined before or during
	 * the compilation of the regular expression, and are used, but
	 * not modified, during text matching.
	 */
	CHAR *cmap;	/* Character mapping array */
	EXPR *root;	/* Pointer to root (EOP) expression */
	unsigned maxid;	/* Number of (also maximum ID of) lit expressions */
	struct BR *br;	/* Backreference pointer */
	EXPR **ptr;	/* Pointer to array of ptrs to lit expressions. */
	BOOL backward;	/* Being matched backwards?  */
	BOOL backref;	/* Backreferencing present? */
	BOOL anchored;	/* Next text char must be matched. */
	int length;	/* Length of matching text, if fixed. 0 if variable */
	int min_length;	/* Min length of matching text. */
	int max_length; /* Max length of matching text. -1 if unlimited. */
	short cost_limit;/* Maximum cost of mismatches for approx. matching. */
	short cost_sub;	/* Cost of substitution mismatch for approx. matching. */
	short cost_del;	/* Cost of deletion mismatch for approx. matching. */
	short cost_ins;	/* Cost of insertion mismatch for approx. matching. */

	/* This following group of attributes are not determined until
	 * after compilation.  Most are variables that can be modified during
	 * text matching itself, although some remain constant after
	 * being initialized after compilation.
	 */
	FID  *posbase;	/* Array of FIDs from all states */
	unsigned nposalloc;/* Allocated size of posbase */
	unsigned posnext;/* Index into free space in posbase */
	unsigned posreset;/* Index into end of IDs for initial state in posbase */
	POSITION_SET ps;/* struct used for calculating state. */
	POSITION_SET fpos;/* struct used for calculating FOLLOW_SET */
	int nstates;	/* Number of current states defined */
	unsigned statelim;/* Limit on number of states before flushing */
	STATE *states;	/* Array of states */
	int flushed;	/* How many times the array of states has been cleared? */
	int threshold;	/* Resize state array every threshold flushes */
} RE_RE;

/* Components of an pattern containing backreferencing. */
typedef enum BR_TYPE {
	br_re,		/* Regular Expression (e.g. "ab.*c*d") or
			 * Group for backreferencing (e.g. "\(..\)")
			 */
	br_br,		/* Backreference string (e.g. "\1") */
	br_brplus,	/* Backreference string, followed by a plus (e.g. "\1+") */
	br_brstar,	/* Backreference string, followed by a star (e.g. "\1*") */
	br_brquest	/* Backreference string, followed by a quest (e.g. "\1?") */
} BR_TYPE;

/* BR:  Structure used for backreferencing. */
typedef struct BR {
	BR_TYPE type;		/* Type of BR subexpression */
	struct GR *group;	/* Pointer to corresponding group */
	RE_RE *r;		/* Pointer to subexpression structure */
	struct BR *br_next;	/* Pointer to BR struct of br_next subexpression */
	CHAR *start;		/* Pointer to text at beginning of last match */
	CHAR *end;		/* Pointer to text at end of last match */
} BR;

/* GR:  Auxiliary structure used for backreferencing. */
typedef struct GR {
	BR *first;	/* Pointer to first BR in group */
	BR *last;	/* Pointer to last BR in group */
	BOOL inside;	/* Are we currently inside this group? */
} GR;


#include "cgre.h"

extern STATE Empty_state;
extern int eg_posalloc(), eg_getstate();
extern void eg_savestate(), eg_stateinit();
extern STATE *nextstate();
extern RE_CW *re_cwinit();
extern void re_cwfreenodes();
extern int re_recw();
extern RE_RE *re_recomp(), *set_count_ptr();
extern void re_cwadd(), re_cwcomp();
extern void re_reinit(), re_brinit(), efollow();
extern BOOL re_cwexec(), re_reexec();
extern BOOL re_cbexec(), re_rbexec();
extern BOOL re_brexec(), re_lnexec(), re_lbexec(), re_nullexec();
extern BOOL interior_nl();
extern CHAR *get_beg();
extern CHAR *cgmalloc(), *cgmalloc2(), *cgcalloc(), *cgrealloc2();
extern CHAR *memchr_fwd(), *memchr_back();
extern EXPR *newexpr();
#endif
