	static char sccsid[]="%Z%	window.c	%E% %U%";
/******************************************************************************

                ###############################################
                # Copyright 1987-2000 Lucent Technologies Inc #
		# Author:  Bill Tanenbaum		      #
                ###############################################

Module:	window.c [%P%]

Description:

	Determines the extent of text windows around matched text.

Functions:
	windows()
	window_limit()
	bound_f_search()
	bound_b_search()
	set_limit()

*******************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "re.h"

static void bound_b_search(), bound_f_search();
static CHAR *set_limit();

/******************************************************************************

Routine:
	windows()

Syntax:

	void windows(pat, bdelim, edelim, blimit, elimit)

	PATTERN *pat;	 // Pointer to patterns that have been matched
	PATTERN *bdelim; // Pointer to begin delimiter to be matched
	PATTERN *edelim; // Pointer to end delimiter to be matched
	CHAR	*blimit; // Pointer to limit of search range for begin delimiter 
	CHAR	*elimit; // Pointer to limit of search range for end delimiter 

Description:

	Delimits the beginning and end of windows by finding matches to the delimiter

	If there were no pattern matches, this routine does nothing.

	Fills in pat->matches[].wbegin for each matched pattern with a pointer
	to the beginning of the window before that match.  However, if
	wbegin != 0, it will assume that the windowing has already been
	done for that match, and not redo it.

	Fills in pat->matches[].wend for each matched pattern with a pointer
	to the end of the window after that match.  However, if
	wend != 0, it will assume that the windowing has already been
	done for that match, and not redo it.

	If the first search for a beginning delimiter runs off the beginning
	of the buffer, and if Check_truncated has been set to YES,
	set Window_truncated to YES.  This means that the beginning of
	a window has been truncated due to the text buffer being too small.

	There is no analogous check in this routine for the truncation of the end
	of a window.  This is because if we run off the end of the buffer, the
	window will not necessarily be truncated.

Global Variables:

	BOOL	Check_truncated;	// Might we be truncating a window?
	BOOL	Window_truncated;	// Truncation has occurred

Files:
	None

Return Status:

	None

*******************************************************************************/

void
windows(pat, bdelim, edelim, blimit, elimit)
register PATTERN *pat, *bdelim, *edelim;
CHAR *blimit, *elimit;
{
	register CGWINDOW *p;
	CHAR *b, *e;

	/* If no matches were found, just return */
	if (!pat->nmatches) {
		return;
	}
	/* Find beginning and end of each window, unless they have been found
	 * previously.
	 */
	for (p = pat->matches; p < pat->matches + pat->nmatches; p++) {
		if (!p->wbegin) {
			/* Beginning of this window not found previously */
			/* Set search boundaries */
			bound_b_search(p, pat, bdelim, blimit, &b, &e);
			/* Search for delimiter */
			p->wbegin = window_limit(bdelim, b, e, 0);
			/* If we are checking for truncation, and this is the
			 * first window in the buffer, and we haven't found
			 * the delimiter, and the search went back to the
			 * beginning of the buffer, then set the flag
			 * indicating that a window has been truncated because
			 * the text buffer was too small.
			 */
			if (Check_truncated && p == pat->matches && bdelim->nmatches == 0 && p->wbegin == blimit) {
				Window_truncated = YES;
			}
		}
		if (!p->wend) {
			/* End of this window not found previously */
			/* Set search boundaries */
			bound_f_search(p, pat, edelim, elimit, &b, &e);
			/* Search for delimiter */
			p->wend = window_limit(edelim, b, e, p->prev_matches);
		}
	}
}

/******************************************************************************

Routine:
	window_limit()

Syntax:

	CHAR *window_limit(delim, b, e, prev_matches)

	PATTERN *delim;	// Pointer to delimiter to be matched
	CHAR *b;	// Pointer to beginning of search range for delimiter 
	CHAR *e;	// Pointer to end of search range for delimiter 
	int prev_matches;// Number of previous matches to the end delimiter
			//    in earlier portions of the window.
			//    This argument is usually 0.  It can be non-zero
			//    only if a window has been split to fit into
			//    the buffer, in which case it is the total number
			//    of matches to the delimiter that have already
			//    occurred in previously scanned portions of the window.
		

Description:

	Searches for a single window delimiter.  If the delimiter was specified,
	returns a pointer to the occurrence of the delimiter that delimits
	the window.

	If the delimiter was not specified, simply return a pointer to the beginning
	of the search range (for the end delimiter) or the end of the search range
	(for the beginning delimiter).

Global Variables:

	None

Files:
	None

Return Status:

	Returns a pointer to the beginning or end of the window.

*******************************************************************************/

CHAR *
window_limit(delim, b, e, prev_matches)
PATTERN *delim;
register CHAR *b, *e;
int prev_matches;
{
	register CHAR *p;

	if (delim->wcount) {
		/* Delimiter specified.  Search for the delimiter,
		 * taking into account any previous matches.
		 */
		delim->N_start = delim->wcount - 1 - prev_matches;
		delim->N_stop = 1;
		delim->nmatches = pattern_match(delim, b, e, 0);
		p = set_limit(delim, (delim->backward ? b : e));
	} else {
		/* No delimiter specified.  The limit of the window is just
		 * the limit of the match.
		 */
		p = (delim->backward ? e : b);
	}
	return(p);
}

/******************************************************************************

Routine:
	bound_f_search()

Syntax:

	static void bound_f_search(p, pat, delim, limit, pb, pe)

	CGWINDOW *p;	// Pointer to window to be delimited	
	PATTERN *pat;	// Pointer to patterns that have been matched
	PATTERN *delim;	// Pointer to delimiter to be matched
	CHAR	*limit;	// Pointer to outside limit of search range for delimiter 
	CHAR	**pb;	// Returns pointer to beginning of search range for delimiter
	CHAR	**pe;	// Returns pointer to end of search range for delimiter

Description:

	Delimits the search range for an end of window delimiter.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
bound_f_search(p, pat, delim, limit, pb, pe)
register CGWINDOW *p;
register PATTERN *pat, *delim;
CHAR *limit;
CHAR **pb, **pe;
{
	/* If no delimiter, the window ends at the end of the match. */
	if (!delim->wcount) {
		*pb = *pe = p->end;
		return;
	}
	/* Set the beginning of the search range to the end of the match */
	*pb = p->end;
	/* If the matched line is itself to be included in the search for
	 * the delimiter, the beginning of the search range must be backed up one
	 * line, but not past the beginning of the buffer or the beginning of
	 * the match.
	 */
	if (delim->incl_match && *pb > Buffer) {
		*pb = memchr_back(--(*pb), NL);
		*pb = MAX(*pb, p->begin);
	}
	/* If the -o flag was used, or if there are no subsequent matches in
	 * the buffer, set the end of the search range to "limit".  Otherwise,
	 * set the end of the search range to the beginning of the next match.
	 */
	if (Flags['o'] || p == pat->matches+pat->nmatches-1) {
		*pe = limit;
	} else {
		*pe = (p+1)->begin;
	}
}

/******************************************************************************

Routine:
	bound_b_search()

Syntax:

	static void bound_b_search(p, pat, delim, limit, pb, pe)

	CGWINDOW *p;	// Pointer to window to be delimited	
	PATTERN *pat;	// Pointer to patterns that have been matched
	PATTERN *delim;	// Pointer to delimiter to be matched
	CHAR	*limit;	// Pointer to outside limit of search range for delimiter 
	CHAR	**pb;	// Returns pointer to beginning of search range for delimiter
	CHAR	**pe;	// Returns pointer to end of search range for delimiter

Description:

	Delimits the search range for a beginning of window delimiter.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/


static void
bound_b_search(p, pat, delim, limit, pb, pe)
register CGWINDOW *p;
register PATTERN *pat, *delim;
CHAR *limit;
CHAR **pb, **pe;
{
	/* If no delimiter, the window begins at the beginning of the match. */
	if (!delim->wcount) {
		*pb = *pe = p->begin;
		return;
	}
	/* Set the end of the search range to the beginning of the match */
	*pe = p->begin;
	/* If the matched line is itself to be included in the search for
	 * the delimiter, the end of the search range must be advanced one
	 * line, but not past the end of the buffer or the end of
	 * the match.
	 */
	if (delim->incl_match && *pe < Buffend) {
		*pe = memchr_fwd(++(*pe), NL);
		*pe = MIN(*pe, p->end);
	}

	/* If the -o flag was used, or if there are no previous matches in
	 * the buffer, set the beginning of the search range to "limit".  Otherwise,
	 * set the beginning of the search range to the end of the previous window.
	 */
	if (Flags['o'] || p == pat->matches) {
		*pb = limit;
	} else {
		*pb = (p-1)->wend;
	}
}

/******************************************************************************

Routine:
	set_limit()

Syntax:

	CHAR * set_limit(delim, p)

	PATTERN *delim;	// Pointer to delimiter that was (or was not) matched
	CHAR	*p;	// Pointer to limit of the search range used for delimiter 

Description:

	Returns a pointer to the beginning or end of the window, after a
	search for the delimiter has occurred.

	If the delimiter was not matched, simply returns "p", the limit
	of the search range.

	If the delimiter was matched, returns a pointer to the beginning
	or end (as appropriate) of the line containing the delimiter.

Global Variables:

	None

Files:
	None

Return Status:

	Returns a pointer to the beginning or end of the window.

*******************************************************************************/

static CHAR *
set_limit(delim, p)
register PATTERN *delim;
register CHAR *p;
{
	if (delim->nmatches) {
		/* The delimiter was matched. */
		if (delim->incl_delim == delim->backward) {
			/* Case 1: Beginning delimiter, and the delimiter is
			 * included in the window.
			 * Case 2: End delimiter, and the delimiter is not
			 * included in the window.
			 * In either case, the window is delimited by the
			 * beginning of the matched delimiter.
			 */
			p = delim->matches[0].begin;
		} else {
			/* Case 3: Beginning delimiter, and the delimiter is
			 * not included in the window.
			 * Case 4: End delimiter, and the delimiter is
			 * included in the window.
			 * In either case, the window is delimited by the
			 * end of the matched delimiter.
			 */
			p = delim->matches[0].end;
		}
	}
	return(p);
}
